local HudRenderTexture = djui_hud_render_texture
local ObjGetFirstWithBehaviorId = obj_get_first_with_behavior_id
local ObjGetNextWithSameBehaviorId = obj_get_next_with_same_behavior_id
local DistBetweenObjectAndPoint = dist_between_object_and_point
local LoadObjectCollisionModel = load_object_collision_model
local HudGetScreenHeight = djui_hud_get_screen_height
local GetTextureInfo = get_texture_info
local ModelUtilGetId = smlua_model_util_get_id
local SpawnSyncObject = spawn_sync_object
local ObjMarkForDeletion = obj_mark_for_deletion
local NetworkIsServer = network_is_server
local MathFloor = math.floor
local MathPi = math.pi
local MathSin = math.sin
local MathCos = math.cos

local _models = {}
_models[0]  = ModelUtilGetId("dirt")
_models[1]  = ModelUtilGetId("grass")
_models[2]  = ModelUtilGetId("carpet")
_models[3]  = ModelUtilGetId("roof")
_models[4]  = ModelUtilGetId("chess")
_models[5]  = ModelUtilGetId("elevator")
_models[6]  = ModelUtilGetId("lantern")
_models[7]  = ModelUtilGetId("metal")
_models[8]  = ModelUtilGetId("mud")
_models[9]  = ModelUtilGetId("pave")
_models[10] = ModelUtilGetId("window")
_models[11] = ModelUtilGetId("stone")
_models[12] = ModelUtilGetId("wall1")
_models[13] = ModelUtilGetId("wall2")
_models[14] = ModelUtilGetId("wall3")
_models[15] = ModelUtilGetId("wall4")
_models[16] = ModelUtilGetId("wall5")
_models[17] = ModelUtilGetId("wall6")
_models[18] = ModelUtilGetId("wall7")
_models[19] = ModelUtilGetId("wall8")

local _textures = {}
_textures[0]  = GetTextureInfo("dirt")
_textures[1]  = GetTextureInfo("grass")
_textures[2]  = GetTextureInfo("carpet")
_textures[3]  = GetTextureInfo("roof")
_textures[4]  = GetTextureInfo("chess")
_textures[5]  = GetTextureInfo("elevator")
_textures[6]  = GetTextureInfo("lantern")
_textures[7]  = GetTextureInfo("metal")
_textures[8]  = GetTextureInfo("mud")
_textures[9]  = GetTextureInfo("pave")
_textures[10] = GetTextureInfo("window")
_textures[11] = GetTextureInfo("stone")
_textures[12] = GetTextureInfo("wall1")
_textures[13] = GetTextureInfo("wall2")
_textures[14] = GetTextureInfo("wall3")
_textures[15] = GetTextureInfo("wall4")
_textures[16] = GetTextureInfo("wall5")
_textures[17] = GetTextureInfo("wall6")
_textures[18] = GetTextureInfo("wall7")
_textures[19] = GetTextureInfo("wall8")

local function BhvBoxInit(obj)
	obj.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
	obj.collisionData = gGlobalObjectCollisionData.breakable_box_seg8_collision_08012D70
	obj.oCollisionDistance = 500
end

local function BhvBoxLoop(obj)
	LoadObjectCollisionModel()
end

_bhvBox = hook_behavior(nil, OBJ_LIST_SURFACE, true, BhvBoxInit, BhvBoxLoop)
local _id = 0

local function GetRelativePos(m)
	local pos = { x = m.pos.x, y = m.pos.y, z = m.pos.z }
	local inAir = m.pos.y - m.floorHeight > 5
	local holdZ = (m.controller.buttonDown & Z_TRIG) ~= 0
	if not inAir then
		local dir = m.faceAngle.y * MathPi * 2 / 65535
		pos.x = pos.x + MathSin(dir) * 200
		pos.z = pos.z + MathCos(dir) * 200
	end
	if inAir or holdZ then
		pos.y = pos.y - 200
	end
	pos.x = MathFloor(pos.x / 200 + 0.5) * 200
	pos.y = MathFloor(pos.y / 200 + 0.5) * 200
	pos.z = MathFloor(pos.z / 200 + 0.5) * 200
	return pos
end

local function FindBlock(pos)
	local check = ObjGetFirstWithBehaviorId(_bhvBox)
	while check ~= nil do
		if DistBetweenObjectAndPoint(check, pos.x, pos.y, pos.z) < 10 then
			return check
		end
		check = ObjGetNextWithSameBehaviorId(check)
	end
end

local function SpawnBlock(pos, id)
	local check = FindBlock(pos)
	if check ~= nil then
		return
	end
	local obj = SpawnSyncObject(
		_bhvBox,
		_models[id],
		pos.x,
		pos.y,
		pos.z,
		function (obj)
			obj.oFaceAnglePitch = 0
			obj.oFaceAngleYaw = 0
			obj.oFaceAngleRoll = 0
		end
	)
end

local function BuildHouse()
	local data = {
		-- left
		                      -400,  800, -200, 15, -400,  800, -400, 15, -400,  800, -600, 15, -400,  800, -800, 15, -400,  800, -1000, 15, -400,  800, -1200, 15, -400,  800, -1400, 15, -400,  800, -1600, 15, -400,  800, -1800, 15, -400,  800, -2000, 15, -400,  800, -2200,  2, -400,  800, -2400,  2, -400,  800, -2600,  2, -400,  800, -2800,  2, -400,  800, -3000,  2,
		                      -400, 1000, -200, 15,                                             -400, 1000, -800, 15,                                               -400, 1000, -1400, 15,                                               -400, 1000, -2000, 15, -400, 1000, -2200,  2, -400, 1000, -2400,  6, -400, 1000, -2600,  2, -400, 1000, -2800,  6, -400, 1000, -3000,  2,
		                      -400, 1200, -200, 15, -400, 1200, -400, 15, -400, 1200, -600, 15, -400, 1200, -800, 15, -400, 1200, -1000, 15, -400, 1200, -1200, 15, -400, 1200, -1400, 15, -400, 1200, -1600, 15, -400, 1200, -1800, 15, -400, 1200, -2000, 15, -400, 1200, -2200,  2, -400, 1200, -2400,  2, -400, 1200, -2600,  2, -400, 1200, -2800,  2, -400, 1200, -3000,  2,
		-- right
		                       400,  800, -200, 15,  400,  800, -400, 15,  400,  800, -600, 15,  400,  800, -800, 15,  400,  800, -1000, 15,  400,  800, -1200, 15,  400,  800, -1400, 15,  400,  800, -1600, 15,  400,  800, -1800, 15,  400,  800, -2000, 15,  400,  800, -2200,  2,  400,  800, -2400,  2,  400,  800, -2600,  2,  400,  800, -2800,  2,  400,  800, -3000,  2,
		                       400, 1000, -200, 15,                                              400, 1000, -800, 15,                                                400, 1000, -1400, 15,                                                400, 1000, -2000, 15,  400, 1000, -2200,  2,  400, 1000, -2400,  6,  400, 1000, -2600,  2,  400, 1000, -2800,  6,  400, 1000, -3000,  2,
		                       400, 1200, -200, 15,  400, 1200, -400, 15,  400, 1200, -600, 15,  400, 1200, -800, 15,  400, 1200, -1000, 15,  400, 1200, -1200, 15,  400, 1200, -1400, 15,  400, 1200, -1600, 15,  400, 1200, -1800, 15,  400, 1200, -2000, 15,  400, 1200, -2200,  2,  400, 1200, -2400,  2,  400, 1200, -2600,  2,  400, 1200, -2800,  2,  400, 1200, -3000,  2,
		-- ceiling
		                                            -200, 1400, -400, 14, -200, 1400, -600, 14, -200, 1400, -800, 14, -200, 1400, -1000, 14, -200, 1400, -1200, 14, -200, 1400, -1400, 14, -200, 1400, -1600, 14, -200, 1400, -1800, 14, -200, 1400, -2000, 14, -200, 1400, -2200, 14, -200, 1400, -2400, 14, -200, 1400, -2600, 14, -200, 1400, -2800, 14, -200, 1400, -3000, 14,
		                                               0, 1400, -400, 14,    0, 1400, -600, 14,    0, 1400, -800, 14,    0, 1400, -1000, 14,    0, 1400, -1200, 14,    0, 1400, -1400, 14,    0, 1400, -1600, 14,    0, 1400, -1800, 14,    0, 1400, -2000, 14,    0, 1400, -2200, 14,    0, 1400, -2400, 14,    0, 1400, -2600, 14,    0, 1400, -2800, 14,    0, 1400, -3000, 14,
		                                             200, 1400, -400, 14,  200, 1400, -600, 14,  200, 1400, -800, 14,  200, 1400, -1000, 14,  200, 1400, -1200, 14,  200, 1400, -1400, 14,  200, 1400, -1600, 14,  200, 1400, -1800, 14,  200, 1400, -2000, 14,  200, 1400, -2200, 14,  200, 1400, -2400, 14,  200, 1400, -2600, 14,  200, 1400, -2800, 14,  200, 1400, -3000, 14,
		-- roof
		-600, 1400,    0,  3, -600, 1400, -200,  3, -600, 1400, -400,  3, -600, 1400, -600,  3, -600, 1400, -800,  3, -600, 1400, -1000,  3, -600, 1400, -1200,  3, -600, 1400, -1400,  3, -600, 1400, -1600,  3, -600, 1400, -1800,  3, -600, 1400, -2000,  3, -600, 1400, -2200,  3,
		-400, 1600,    0,  3, -400, 1600, -200,  3, -400, 1600, -400,  3, -400, 1600, -600,  3, -400, 1600, -800,  3, -400, 1600, -1000,  3, -400, 1600, -1200,  3, -400, 1600, -1400,  3, -400, 1600, -1600,  3, -400, 1600, -1800,  3, -400, 1600, -2000,  3, -400, 1600, -2200,  3,
		-200, 1800,    0,  3, -200, 1800, -200,  3, -200, 1800, -400,  3, -200, 1800, -600,  3, -200, 1800, -800,  3, -200, 1800, -1000,  3, -200, 1800, -1200,  3, -200, 1800, -1400,  3, -200, 1800, -1600,  3, -200, 1800, -1800,  3, -200, 1800, -2000,  3, -200, 1800, -2200,  3,
		   0, 2000,    0,  3,    0, 2000, -200,  3,    0, 2000, -400,  3,    0, 2000, -600,  3,    0, 2000, -800,  3,    0, 2000, -1000,  3,    0, 2000, -1200,  3,    0, 2000, -1400,  3,    0, 2000, -1600,  3,    0, 2000, -1800,  3,    0, 2000, -2000,  3,    0, 2000, -2200,  3,
		 200, 1800,    0,  3,  200, 1800, -200,  3,  200, 1800, -400,  3,  200, 1800, -600,  3,  200, 1800, -800,  3,  200, 1800, -1000,  3,  200, 1800, -1200,  3,  200, 1800, -1400,  3,  200, 1800, -1600,  3,  200, 1800, -1800,  3,  200, 1800, -2000,  3,  200, 1800, -2200,  3,
		 400, 1600,    0,  3,  400, 1600, -200,  3,  400, 1600, -400,  3,  400, 1600, -600,  3,  400, 1600, -800,  3,  400, 1600, -1000,  3,  400, 1600, -1200,  3,  400, 1600, -1400,  3,  400, 1600, -1600,  3,  400, 1600, -1800,  3,  400, 1600, -2000,  3,  400, 1600, -2200,  3,
		 600, 1400,    0,  3,  600, 1400, -200,  3,  600, 1400, -400,  3,  600, 1400, -600,  3,  600, 1400, -800,  3,  600, 1400, -1000,  3,  600, 1400, -1200,  3,  600, 1400, -1400,  3,  600, 1400, -1600,  3,  600, 1400, -1800,  3,  600, 1400, -2000,  3,  600, 1400, -2200,  3,
		-- others
		-400, 1400, -200, 15, -200, 1400, -200, 15, 0, 1400, -200, 15,  200, 1400, -200, 15, 400, 1400, -200, 15, -200, 1600, -200, 15, 0, 1600, -200, 15,  200, 1600, -200, 15, 0, 1800, -200, 15,
		-600,  800, -2200, 15, -600, 1000, -2200, 15, -600, 1200, -2200, 15, 600,  800, -2200, 15, 600, 1000, -2200, 15, 600, 1200, -2200, 15,
	}
	for i = 1, #data, 4 do
		local pos = { x = data[i], y = data[i + 1], z = data[i + 2] }
		SpawnBlock(pos, data[i + 3])
	end
end

local function MarioUpdate(m)
	if m.playerIndex == 0 and (m.controller.buttonPressed & R_JPAD) ~= 0 then
		_id = _id + 1
		if _id > #_textures then
			_id = 0
		end
	elseif m.playerIndex == 0 and (m.controller.buttonPressed & L_JPAD) ~= 0 then
		_id = _id - 1
		if _id < 0 then
			_id = #_textures
		end
	elseif m.playerIndex == 0 and (m.controller.buttonPressed & X_BUTTON) ~= 0 then
		local pos = GetRelativePos(m)
		local check = FindBlock(pos)
		if check ~= nil then
			network_send(false, { x = check.oPosX, y = check.oPosY, z = check.oPosZ } )
			ObjMarkForDeletion(check)
		end
	elseif m.playerIndex == 0 and (m.controller.buttonPressed & Y_BUTTON) ~= 0 then
		local pos = GetRelativePos(m)
		SpawnBlock(pos, _id)
	end
end

local function ChatCommand(msg)
	BuildHouse()
	return true
end

function OnPacketReceive(pos)
	local check = FindBlock(pos)
	if check ~= nil then
		ObjMarkForDeletion(check)
	end
end

function OnHudRender()
	local texture = _textures[_id]
	local scale = HudGetScreenHeight() / texture.height / 8
	local top = HudGetScreenHeight() - texture.height * scale
	HudRenderTexture(texture, 0, top, scale, scale)
end

hook_event(HOOK_ON_PACKET_RECEIVE, OnPacketReceive)
hook_event(HOOK_ON_HUD_RENDER, OnHudRender)
hook_event(HOOK_MARIO_UPDATE, MarioUpdate)
hook_chat_command("house", "house", ChatCommand)
